@@ -20,7 +20,7 @@ from rasa.shared.constants import DOCS_BASE_URL, DEFAULT_SENDER_ID
 from rasa.shared.exceptions import RasaException
 
 try:
-    from urlparse import urljoin  # pytype: disable=import-error
+    from urlparse import urljoin
 except ImportError:
     from urllib.parse import urljoin
 
@@ -319,7 +319,7 @@ class CollectingOutputChannel(OutputChannel):
             return None
 
     async def _persist_message(self, message: Dict[Text, Any]) -> None:
-        self.messages.append(message)  # pytype: disable=bad-return-type
+        self.messages.append(message)
 
     async def send_text_message(
         self, recipient_id: Text, text: Text, **kwargs: Any
