@@ -476,7 +476,8 @@ class MessageProcessor:
         input_channel = tracker.get_latest_input_channel()
 
         tracker.update(
-            UserUttered.create_external(intent_name, entity_list, input_channel)
+            UserUttered.create_external(intent_name, entity_list, input_channel),
+            self.domain,
         )
         await self._predict_and_execute_next_action(output_channel, tracker)
         # save tracker state to continue conversation from this state
